@@ -1,4 +1,3 @@
-use crate::Instr;
 use crate::error::Error;
 use crate::partial_execution::*;
 use crate::smallvm::ParsedInstr;
@@ -8,27 +7,21 @@ use crossbeam::channel::unbounded;
 
 use log::trace;
 
-use petgraph::algo::astar;
-use petgraph::algo::dijkstra;
+use petgraph::algo::{astar, dijkstra};
 use petgraph::graph::{EdgeIndex, Graph, NodeIndex};
 use petgraph::visit::{Bfs, EdgeRef};
-use petgraph::Direction;
-use petgraph::IntoWeightedEdge;
-use py27_marshal::bstr::ByteSlice;
+use petgraph::{Direction, IntoWeightedEdge};
 use py27_marshal::{Code, Obj};
 use pydis::prelude::*;
+use pydis::opcode::py27::{self, Mnemonic};
 use std::fmt;
 
+use std::collections::{BTreeSet, HashMap};
 use std::fs::File;
 use std::io::Write;
-use std::sync::atomic::Ordering;
+use std::marker::PhantomData;
+use std::sync::atomic::{AtomicUsize, Ordering};
 use std::sync::{Arc, Mutex};
-use std::{
-    collections::{BTreeSet, HashMap},
-    sync::atomic::AtomicUsize,
-};
-
-type TargetOpcode = pydis::opcode::Python27;
 
 bitflags! {
     #[derive(Default)]
@@ -68,14 +61,14 @@ impl fmt::Display for EdgeWeight {
 }
 
 /// Represents a single block of code up until its next branching point
-#[derive(Debug, Default)]
-pub struct BasicBlock {
+#[derive(Debug)]
+pub struct BasicBlock<O: Opcode<Mnemonic = py27::Mnemonic>> {
     /// Offset of the first instruction in this BB
     pub start_offset: u64,
     /// Offset of the last instruction in this BB (note: this is the START of the last instruction)
     pub end_offset: u64,
     /// Instructions contained within this BB
-    pub instrs: Vec<ParsedInstr>,
+    pub instrs: Vec<ParsedInstr<O>>,
     /// Whether this BB contains invalid instructions
     pub has_bad_instrs: bool,
     /// Flags used for internal purposes
@@ -84,7 +77,20 @@ pub struct BasicBlock {
     last_instruction_end: u64,
 }
 
-impl fmt::Display for BasicBlock {
+impl<O: Opcode<Mnemonic = py27::Mnemonic>> Default for BasicBlock<O> {
+    fn default() -> Self {
+        BasicBlock {
+            start_offset: 0,
+            end_offset: 0,
+            instrs: vec![],
+            has_bad_instrs: false,
+            flags: BasicBlockFlags::default(),
+            last_instruction_end: 0,
+        }
+    }
+}
+
+impl<O: Opcode<Mnemonic = py27::Mnemonic>> fmt::Display for BasicBlock<O> {
     fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
         writeln!(f, "Flags: {:?}", self.flags)?;
         let mut offset = self.start_offset;
@@ -104,10 +110,10 @@ impl fmt::Display for BasicBlock {
     }
 }
 
-impl BasicBlock {
+impl<O: Opcode<Mnemonic = py27::Mnemonic>> BasicBlock<O> {
     /// Splits a basic block at the target absolute offset. The instruction index is calculated
     /// on-demand, walking the instructions and adding their length until the desired offset is found.
-    pub fn split(&mut self, offset: u64) -> Option<(u64, BasicBlock)> {
+    pub fn split(&mut self, offset: u64) -> Option<(u64, BasicBlock<O>)> {
         // It does indeed land in the middle of this block. Let's figure out which
         // instruction it lands on
         let mut ins_offset = self.start_offset;
@@ -149,25 +155,26 @@ impl BasicBlock {
 }
 
 /// A code object represented as a graph
-pub struct CodeGraph {
+pub struct CodeGraph<TargetOpcode: Opcode<Mnemonic = py27::Mnemonic>> {
     pub(crate) root: NodeIndex,
     code: Arc<Code>,
-    pub(crate) graph: Graph<BasicBlock, EdgeWeight>,
+    pub(crate) graph: Graph<BasicBlock<TargetOpcode>, EdgeWeight>,
     file_identifier: usize,
     /// Whether or not to generate dotviz graphs
     enable_dotviz_graphs: bool,
     phase: usize,
     /// Hashmap of graph file names and their data
     pub(crate) dotviz_graphs: HashMap<String, String>,
+    _target_opcode_phantom: PhantomData<TargetOpcode>,
 }
 
-impl CodeGraph {
+impl<TargetOpcode: 'static + Opcode<Mnemonic = py27::Mnemonic>> CodeGraph<TargetOpcode> {
     /// Converts bytecode to a graph. Returns the root node index and the graph.
     pub fn from_code(
         code: Arc<Code>,
         file_identifier: usize,
         enable_dotviz_graphs: bool,
-    ) -> Result<CodeGraph, Error> {
+    ) -> Result<CodeGraph<TargetOpcode>, Error<TargetOpcode>> {
         let debug = false;
 
         let analyzed_instructions = crate::smallvm::const_jmp_instruction_walker(
@@ -185,7 +192,7 @@ impl CodeGraph {
         }
 
         let mut curr_basic_block = BasicBlock::default();
-        let mut code_graph = petgraph::Graph::<BasicBlock, EdgeWeight>::new();
+        let mut code_graph = petgraph::Graph::<BasicBlock<TargetOpcode>, EdgeWeight>::new();
         let mut edges = vec![];
         let mut root_node_id = None;
         let mut has_invalid_jump_sites = false;
@@ -229,8 +236,8 @@ impl CodeGraph {
             };
 
             if matches!(
-                instr.opcode,
-                TargetOpcode::RETURN_VALUE | TargetOpcode::RAISE_VARARGS
+                instr.opcode.mnemonic(),
+                Mnemonic::RETURN_VALUE | Mnemonic::RAISE_VARARGS
             ) {
                 curr_basic_block.end_offset = offset;
                 // We need to see if a previous BB landed in the middle of this block.
@@ -304,8 +311,8 @@ impl CodeGraph {
                         matches!(&copy.get(&target), Some(ParsedInstr::Bad) | None);
 
                     let edge_weight = if matches!(
-                        instr.opcode,
-                        TargetOpcode::JUMP_FORWARD | TargetOpcode::JUMP_ABSOLUTE,
+                        instr.opcode.mnemonic(),
+                        Mnemonic::JUMP_FORWARD | Mnemonic::JUMP_ABSOLUTE,
                     ) {
                         EdgeWeight::NonJump
                     } else {
@@ -427,6 +434,7 @@ impl CodeGraph {
             enable_dotviz_graphs,
             phase: 0,
             dotviz_graphs: HashMap::new(),
+            _target_opcode_phantom: Default::default(),
         })
     }
 
@@ -434,7 +442,8 @@ impl CodeGraph {
     /// $FILENUMBER_$FILENAME_$NAME_$STAGE.dot. This function will check global args to see if
     /// dot writing was enabled
     pub fn generate_dot_graph(&mut self, stage: &str) {
-        if !self.enable_dotviz_graphs {
+        let force_graphs = std::env::var("UNFUCK_WRITE_GRAPHS").is_ok();
+        if !self.enable_dotviz_graphs && !force_graphs {
             return;
         }
 
@@ -451,15 +460,14 @@ impl CodeGraph {
 
         self.phase += 1;
 
-        let dot_data= format!("{}", Dot::with_config(&self.graph, &[Config::EdgeNoLabel]));
-        if std::env::var("UNFUCK_WRITE_GRAPHS").is_ok() {
+        let dot_data = format!("{}", Dot::with_config(&self.graph, &[Config::EdgeNoLabel]));
+        if force_graphs {
             let mut output_file = File::create(&filename).expect("failed to create dot file");
-            output_file.write_all(dot_data.as_bytes()).expect("failed to write dot data");
+            output_file
+                .write_all(dot_data.as_bytes())
+                .expect("failed to write dot data");
         }
-        self.dotviz_graphs.insert(
-            filename,
-            dot_data
-        );
+        self.dotviz_graphs.insert(filename, dot_data);
     }
 
     fn invoke_partial_execution(
@@ -854,7 +862,7 @@ impl CodeGraph {
         while let Some(nx) = bfs.next(&self.graph) {
             let bb = &self.graph[nx];
             if let Some(instr) = bb.instrs.first() {
-                if instr.unwrap().opcode == TargetOpcode::FOR_ITER {
+                if instr.unwrap().opcode.mnemonic() == Mnemonic::FOR_ITER {
                     // let's get this target node -- if it's just a RETURN_VALUE with other
                     // people returning as well, we should just use our own
                     //
@@ -882,7 +890,7 @@ impl CodeGraph {
 
                     let target_bb = &self.graph[target];
                     if target_bb.instrs.len() == 1
-                        && target_bb.instrs[0].unwrap().opcode == TargetOpcode::RETURN_VALUE
+                        && target_bb.instrs[0].unwrap().opcode.mnemonic() == Mnemonic::RETURN_VALUE
                         && self
                             .graph
                             .edges_directed(target, Direction::Incoming)
@@ -892,7 +900,7 @@ impl CodeGraph {
                         self.graph.remove_edge(edge);
                         let return_bb = BasicBlock {
                             instrs: vec![crate::smallvm::ParsedInstr::Good(Arc::new(
-                                pydis::Instr!(TargetOpcode::RETURN_VALUE),
+                                pydis::Instr!(Mnemonic::RETURN_VALUE.into()),
                             ))],
                             ..Default::default()
                         };
@@ -949,18 +957,18 @@ impl CodeGraph {
                 let source_node = &mut self.graph[incoming_edge];
                 let end_of_jump_ins = source_node.end_offset + last_ins_len as u64;
 
-                if last_ins.opcode == TargetOpcode::JUMP_ABSOLUTE
+                if last_ins.opcode.mnemonic() == Mnemonic::JUMP_ABSOLUTE
                     && target_node_start > source_node.start_offset
                 {
                     unsafe { Arc::get_mut_unchecked(&mut last_ins) }.opcode =
-                        TargetOpcode::JUMP_FORWARD;
+                        Mnemonic::JUMP_FORWARD.into();
                 }
 
-                if last_ins.opcode == TargetOpcode::JUMP_FORWARD
+                if last_ins.opcode.mnemonic() == Mnemonic::JUMP_FORWARD
                     && target_node_start < end_of_jump_ins
                 {
                     unsafe { Arc::get_mut_unchecked(&mut last_ins) }.opcode =
-                        TargetOpcode::JUMP_ABSOLUTE;
+                        Mnemonic::JUMP_ABSOLUTE.into();
                 }
 
                 let last_ins_is_abs_jump = last_ins.opcode.is_absolute_jump();
@@ -1009,7 +1017,7 @@ impl CodeGraph {
             current_node.flags |= BasicBlockFlags::BYTECODE_WRITTEN;
 
             for instr in current_node.instrs.iter().map(|i| i.unwrap()) {
-                new_bytecode.push(instr.opcode as u8);
+                new_bytecode.push(instr.opcode.to_u8().unwrap());
                 if let Some(arg) = instr.arg {
                     new_bytecode.extend_from_slice(&arg.to_le_bytes()[..]);
                 }
@@ -1128,8 +1136,8 @@ impl CodeGraph {
                 for instr in &self.graph[node].instrs {
                     // these ones pop only if we're not taking the branch
                     if matches!(
-                        instr.unwrap().opcode,
-                        TargetOpcode::JUMP_IF_TRUE_OR_POP | TargetOpcode::JUMP_IF_FALSE_OR_POP
+                        instr.unwrap().opcode.mnemonic(),
+                        Mnemonic::JUMP_IF_TRUE_OR_POP | Mnemonic::JUMP_IF_FALSE_OR_POP
                     ) {
                         // Grab the edge from this node to the next
                         let edge = self.graph.find_edge(node, path[idx + 1]).unwrap();
@@ -1140,8 +1148,8 @@ impl CodeGraph {
                         }
                     } else {
                         if matches!(
-                            instr.unwrap().opcode,
-                            TargetOpcode::SETUP_EXCEPT | TargetOpcode::SETUP_FINALLY
+                            instr.unwrap().opcode.mnemonic(),
+                            Mnemonic::SETUP_EXCEPT | Mnemonic::SETUP_FINALLY
                         ) {
                             stack_size += 0;
                         } else {
@@ -1157,7 +1165,7 @@ impl CodeGraph {
                 current_node
                     .instrs
                     .push(ParsedInstr::Good(Arc::new(Instruction {
-                        opcode: TargetOpcode::POP_TOP,
+                        opcode: Mnemonic::POP_TOP.into(),
                         arg: None,
                     })));
             }
@@ -1178,13 +1186,13 @@ impl CodeGraph {
             current_node
                 .instrs
                 .push(ParsedInstr::Good(Arc::new(Instruction {
-                    opcode: TargetOpcode::LOAD_CONST,
+                    opcode: Mnemonic::LOAD_CONST.into(),
                     arg: Some(const_idx as u16),
                 })));
             current_node
                 .instrs
                 .push(ParsedInstr::Good(Arc::new(Instruction {
-                    opcode: TargetOpcode::RETURN_VALUE,
+                    opcode: Mnemonic::RETURN_VALUE.into(),
                     arg: None,
                 })));
 
@@ -1252,7 +1260,7 @@ impl CodeGraph {
                     self.graph[nx]
                         .instrs
                         .push(ParsedInstr::Good(Arc::new(pydis::Instr!(
-                            TargetOpcode::JUMP_FORWARD,
+                            Mnemonic::JUMP_FORWARD.into(),
                             0
                         ))));
                 }
@@ -1277,7 +1285,7 @@ impl CodeGraph {
                                 if *outstanding == target {
                                     outstanding_conditions.pop();
                                     self.graph[nx].instrs.push(ParsedInstr::Good(Arc::new(
-                                        pydis::Instr!(TargetOpcode::JUMP_FORWARD, 0),
+                                        pydis::Instr!(Mnemonic::JUMP_FORWARD.into(), 0),
                                     )));
                                 }
                             }
@@ -1457,10 +1465,14 @@ impl CodeGraph {
                 let removed_instruction = parent_node.instrs.pop().unwrap();
 
                 trace!("{:?}", removed_instruction);
-                //assert!(!removed_instruction.unwrap().opcode.is_conditional_jump(), "Removed instruction is a conditional jump: {:#x?}", removed_instruction);
-                parent_node.instrs.push(ParsedInstr::Good(Arc::new(Instr!(TargetOpcode::POP_TOP))));
-                current_end_offset -= removed_instruction.unwrap().len() as u64;
-                current_end_offset += parent_node.instrs.last().unwrap().unwrap().len() as u64;
+                assert!(
+                    !removed_instruction.unwrap().opcode.is_conditional_jump(),
+                    "Removed instruction is a conditional jump: {:#x?}",
+                    removed_instruction
+                );
+                // parent_node.instrs.push(ParsedInstr::Good(Arc::new(Instr!(TargetOpcode::POP_TOP))));
+                // current_end_offset -= removed_instruction.unwrap().len() as u64;
+                // current_end_offset += parent_node.instrs.last().unwrap().unwrap().len() as u64;
             }
         }
 
@@ -1530,19 +1542,17 @@ pub(crate) mod tests {
     use std::collections::BTreeMap;
 
     use super::*;
-    use crate::deobfuscate_codeobj as main_deob;
     use crate::smallvm::tests::*;
-    use crate::Instr;
+    use crate::{deobfuscate_codeobj as main_deob, Instr};
     use pydis::opcode::Instruction;
 
-    type TargetOpcode = pydis::opcode::Python27;
+    type TargetOpcode = pydis::opcode::py27::Standard;
 
-    fn deobfuscate_codeobj(data: &[u8]) -> Result<Vec<Vec<u8>>, Error> {
+    fn deobfuscate_codeobj(data: &[u8]) -> Result<Vec<Vec<u8>>, Error<TargetOpcode>> {
         let files_processed = AtomicUsize::new(0);
-        main_deob(data, &files_processed, false).map(|res| {
+        main_deob::<TargetOpcode>(data, &files_processed, false).map(|res| {
             let mut output = vec![];
-            let mut code_objects =
-                vec![py27_marshal::read::marshal_loads(data).unwrap()];
+            let mut code_objects = vec![py27_marshal::read::marshal_loads(data).unwrap()];
 
             let mut files_processed = 0;
             while let Some(py27_marshal::Obj::Code(obj)) = code_objects.pop() {
@@ -1674,7 +1684,7 @@ pub(crate) mod tests {
 
         change_code_instrs(&mut code, &instrs[..]);
 
-        let mut code_graph = CodeGraph::from_code(code, 0, false).unwrap();
+        let mut code_graph = CodeGraph::<TargetOpcode>::from_code(code, 0, false).unwrap();
 
         code_graph.join_blocks();
         code_graph.update_bb_offsets();
@@ -1820,8 +1830,8 @@ pub(crate) mod tests {
 
     #[test]
     fn deobfuscate_known_file_compileall() {
-        let obfuscated = include_bytes!("../test_data/obfuscated/compiler/compileall_stage4.pyc");
-        let source_of_truth = vec![]; // TODO: include_bytes!("../test_data/expected/compiler/compileall.pyc");
+        let obfuscated = include_bytes!("../test_data/obfuscated/compileall_stage4.pyc");
+        let source_of_truth = include_bytes!("../test_data/expected/compileall.pyc");
 
         let deobfuscated = deobfuscate_codeobj(&obfuscated[8..]).expect("failed to deobfuscate");
 
@@ -1832,8 +1842,8 @@ pub(crate) mod tests {
         let mut files_processed = 0;
         while let Some(py27_marshal::Obj::Code(obj)) = code_objects.pop() {
             let mut code_graph =
-                CodeGraph::from_code(Arc::clone(&obj), files_processed, false).unwrap();
-                // for debugging
+                CodeGraph::<TargetOpcode>::from_code(Arc::clone(&obj), files_processed, false).unwrap();
+            // for debugging
             code_graph.generate_dot_graph("compileall");
             files_processed += 1;
             source_of_truth_bytecode.push(obj.code.as_ref().clone());
