@@ -1,14 +1,14 @@
 use log::{debug, trace};
-use num_bigint::ToBigInt;
+use num_bigint::{BigInt, ToBigInt};
 use num_traits::{Pow, ToPrimitive};
 use py27_marshal::bstr::BString;
 use py27_marshal::*;
+use pydis::opcode::py27::{self, Mnemonic};
 use pydis::prelude::*;
 use std::collections::{BTreeMap, HashMap, VecDeque};
-use std::convert::TryFrom;
+use std::convert::{TryFrom, TryInto};
 use std::io::{Cursor, Read};
 use std::sync::{Arc, Mutex};
-type TargetOpcode = pydis::opcode::Python27;
 
 pub enum WalkerState {
     /// Continue parsing normally
@@ -127,8 +127,8 @@ pub(crate) const PYTHON27_COMPARE_OPS: [&str; 12] = [
 /// Executes an instruction, altering the input state and returning an error
 /// when the instruction cannot be correctly emulated. For example, some complex
 /// instructions are not currently supported at this time.
-pub fn execute_instruction<F, T>(
-    instr: &Instruction<TargetOpcode>,
+pub fn execute_instruction<O: Opcode<Mnemonic = py27::Mnemonic>, F, T>(
+    instr: &Instruction<O>,
     code: Arc<Code>,
     stack: &mut VmStack<T>,
     vars: &mut VmVars<T>,
@@ -137,7 +137,7 @@ pub fn execute_instruction<F, T>(
     names_loaded: LoadedNames,
     mut function_callback: F,
     access_tracking: T,
-) -> Result<(), Error>
+) -> Result<(), Error<O>>
 where
     F: FnMut(VmVar, Vec<VmVar>, std::collections::HashMap<Option<ObjHashable>, VmVar>) -> VmVar,
     T: Clone + Copy,
@@ -479,8 +479,8 @@ where
         };
     }
 
-    match instr.opcode {
-        TargetOpcode::ROT_TWO => {
+    match instr.opcode.mnemonic() {
+        Mnemonic::ROT_TWO => {
             let (tos, tos_accesses) = stack.pop().unwrap();
             let (tos1, tos1_accesses) = stack.pop().unwrap();
             tos_accesses.push(access_tracking);
@@ -489,7 +489,7 @@ where
             stack.push((tos1, tos1_accesses));
             stack.push((tos, tos_accesses));
         }
-        TargetOpcode::ROT_THREE => {
+        Mnemonic::ROT_THREE => {
             let (tos, tos_accesses) = stack.pop().unwrap();
             let (tos1, tos1_accesses) = stack.pop().unwrap();
             let (tos2, tos2_accesses) = stack.pop().unwrap();
@@ -501,13 +501,13 @@ where
             stack.push((tos1, tos1_accesses));
             stack.push((tos, tos_accesses));
         }
-        TargetOpcode::DUP_TOP => {
+        Mnemonic::DUP_TOP => {
             let (var, accesses) = stack.last().unwrap();
             accesses.push(access_tracking);
             let new_var = (var.clone(), accesses.deep_clone());
             stack.push(new_var);
         }
-        TargetOpcode::COMPARE_OP => {
+        Mnemonic::COMPARE_OP => {
             let (right, right_modifying_instrs) = stack.pop().unwrap();
             let (left, left_modifying_instrs) = stack.pop().unwrap();
 
@@ -775,7 +775,7 @@ where
                 other => panic!("unsupported comparison operator: {:?}", other),
             }
         }
-        TargetOpcode::IMPORT_NAME => {
+        Mnemonic::IMPORT_NAME => {
             let (_fromlist, fromlist_modifying_instrs) = stack.pop().unwrap();
             let (_level, level_modifying_instrs) = stack.pop().unwrap();
 
@@ -787,14 +787,14 @@ where
 
             stack.push((None, level_modifying_instrs));
         }
-        TargetOpcode::IMPORT_FROM => {
+        Mnemonic::IMPORT_FROM => {
             let (_module, accessing_instrs) = stack.last().unwrap();
             accessing_instrs.push(access_tracking);
             let accessing_instrs = accessing_instrs.clone();
 
             stack.push((None, accessing_instrs));
         }
-        TargetOpcode::LOAD_ATTR => {
+        Mnemonic::LOAD_ATTR => {
             // we don't support attributes
             let (_obj, obj_modifying_instrs) = stack.pop().unwrap();
             let _name = &code.names[instr.arg.unwrap() as usize];
@@ -803,12 +803,12 @@ where
 
             stack.push((None, obj_modifying_instrs));
         }
-        TargetOpcode::STORE_ATTR => {
+        Mnemonic::STORE_ATTR => {
             // we don't support attributes
             let (_obj, _obj_modifying_instrs) = stack.pop().unwrap();
             let (_obj, _obj_modifying_instrs) = stack.pop().unwrap();
         }
-        TargetOpcode::FOR_ITER => {
+        Mnemonic::FOR_ITER => {
             // Top of stack needs to be something we can iterate over
             // get the next item from our iterator
             let top_of_stack_index = stack.len() - 1;
@@ -831,20 +831,20 @@ where
 
             stack.push((new_tos, InstructionTracker::new()))
         }
-        TargetOpcode::STORE_FAST => {
+        Mnemonic::STORE_FAST => {
             let (tos, accessing_instrs) = stack.pop().unwrap();
             accessing_instrs.push(access_tracking);
             // Store TOS in a var slot
             vars.insert(instr.arg.unwrap(), (tos, accessing_instrs));
         }
-        TargetOpcode::STORE_NAME => {
+        Mnemonic::STORE_NAME => {
             let (tos, accessing_instrs) = stack.pop().unwrap();
             let name = &code.names[instr.arg.unwrap() as usize];
             accessing_instrs.push(access_tracking);
             // Store TOS in a var slot
             names.insert(Arc::clone(name), (tos, accessing_instrs));
         }
-        TargetOpcode::LOAD_NAME => {
+        Mnemonic::LOAD_NAME => {
             let name = &code.names[instr.arg.unwrap() as usize];
             names_loaded.lock().unwrap().push(Arc::clone(name));
             if let Some((val, accesses)) = names.get(name) {
@@ -856,7 +856,7 @@ where
                 stack.push((None, tracking));
             }
         }
-        TargetOpcode::LOAD_FAST => {
+        Mnemonic::LOAD_FAST => {
             if let Some((var, accesses)) = vars.get(&instr.arg.unwrap()) {
                 accesses.push(access_tracking);
                 stack.push((var.clone(), accesses.clone()));
@@ -866,7 +866,7 @@ where
                 stack.push((None, tracking));
             }
         }
-        TargetOpcode::LOAD_CONST => {
+        Mnemonic::LOAD_CONST => {
             let tracking = InstructionTracker::new();
             tracking.push(access_tracking);
 
@@ -875,76 +875,70 @@ where
                 tracking,
             ));
         }
-        TargetOpcode::BINARY_FLOOR_DIVIDE => {
+        Mnemonic::BINARY_FLOOR_DIVIDE => {
             apply_operator!("//");
         }
-        TargetOpcode::BINARY_TRUE_DIVIDE => {
+        Mnemonic::BINARY_TRUE_DIVIDE => {
             apply_operator!("///");
         }
-        TargetOpcode::BINARY_POWER => {
+        Mnemonic::BINARY_POWER => {
             apply_operator!("**");
         }
-        TargetOpcode::BINARY_MODULO => {
+        Mnemonic::BINARY_MODULO => {
             apply_operator!("%");
         }
-        TargetOpcode::INPLACE_ADD | TargetOpcode::BINARY_ADD => {
+        Mnemonic::INPLACE_ADD | Mnemonic::BINARY_ADD => {
             apply_operator!("+");
         }
-        TargetOpcode::INPLACE_MULTIPLY | TargetOpcode::BINARY_MULTIPLY => {
+        Mnemonic::INPLACE_MULTIPLY | Mnemonic::BINARY_MULTIPLY => {
             apply_operator!("*");
         }
-        TargetOpcode::INPLACE_SUBTRACT | TargetOpcode::BINARY_SUBTRACT => {
+        Mnemonic::INPLACE_SUBTRACT | Mnemonic::BINARY_SUBTRACT => {
             apply_operator!("-");
         }
-        TargetOpcode::STORE_SUBSCR => {
-            return Err(
-                crate::error::ExecutionError::ComplexExpression(instr.clone(), None).into(),
-            );
-            let (_tos, _accessing_instrs) = stack.pop().unwrap();
-            let (_tos1, _tos1_accessing_instrs) = stack.pop().unwrap();
-            let (_tos2, _tos2_accessing_instrs) = stack.pop().unwrap();
-            return Ok(());
-            // accessing_instrs
-            //     .borrow_mut()
-            //     .extend_from_slice(tos1_accessing_instrs.borrow().as_slice());
-            // accessing_instrs
-            //     .borrow_mut()
-            //     .extend_from_slice(tos2_accessing_instrs.borrow().as_slice());
-            // accessing_instrs.borrow_mut().push(access_tracking);
+        Mnemonic::STORE_SUBSCR => {
+            let (key, key_accessing_instrs) = stack.pop().unwrap();
+            let (collection, collection_accessing_instrs) = stack.pop().unwrap();
+            let (value, value_accessing_instrs) = stack.pop().unwrap();
 
-            // if tos.is_none() || tos2.is_none() {
-            //     match tos1 {
-            //         Some(Obj::Dict(list_lock)) => {
-            //             let mut dict = list_lock.write().unwrap();
-            //             let key = ObjHashable::try_from(&tos).unwrap();
-            //             dict.insert(key, tos2);
-            //         }
-            //         Some(other) => {
-            //             panic!("need to implement BINARY_SUBSC for set");
-            //         }
-            //         None => {
-            //             stack.push((None, accessing_instrs));
-            //         }
-            //     }
-            // }
-            // let tos = tos.unwrap();
-            // let tos2 = tos2.unwrap();
+            collection_accessing_instrs.extend(&key_accessing_instrs);
 
-            // match tos1 {
-            //     Some(Obj::Dict(list_lock)) => {
-            //         let mut dict = list_lock.write().unwrap();
-            //         let key = ObjHashable::try_from(&tos).unwrap();
-            //         dict.insert(key, tos2);
-            //     }
-            //     Some(other) => {
-            //         panic!("need to implement BINARY_SUBSC for set");
-            //     }
-            //     None => {
-            //         stack.push((None, accessing_instrs));
-            //     }
-            // }
+            collection_accessing_instrs.extend(&value_accessing_instrs);
+
+            collection_accessing_instrs.push(access_tracking);
+
+            // If key, value, or the collection are `None`, we destroy the entire collection
+            // TODO: allow more granular failure of taint tracking at a per-index level
+
+            match (collection, key, value) {
+                (Some(collection), Some(key), Some(value)) => match collection {
+                    Obj::Dict(list_lock) => {
+                        let mut dict = list_lock.write().unwrap();
+                        let key = ObjHashable::try_from(&key).expect("key is not hashable");
+                        dict.insert(key, value);
+                    }
+                    Obj::List(list_lock) => {
+                        let mut list = list_lock.write().unwrap();
+                        let index = key.extract_long().expect("key is not a long");
+                        let index = index
+                            .to_usize()
+                            .expect("index cannot be converted to usize");
+                        if index > list.len() {
+                            panic!("index {} is greater than list length {}", index, list.len());
+                        }
+
+                        list[index] = value;
+                    }
+                    other => {
+                        panic!("need to implement STORE_SUBSCR for {:?}", other.typ());
+                    }
+                },
+                _ => {
+                    // we do nothing
+                }
+            }
         }
-        TargetOpcode::BINARY_SUBSC => {
+        Mnemonic::BINARY_SUBSC => {
             let (tos, accessing_instrs) = stack.pop().unwrap();
             let (tos1, tos1_accessing_instrs) = stack.pop().unwrap();
             accessing_instrs.extend(&tos1_accessing_instrs);
@@ -983,25 +977,25 @@ where
                 }
             }
         }
-        TargetOpcode::BINARY_DIVIDE => {
+        Mnemonic::BINARY_DIVIDE => {
             apply_operator!("/");
         }
-        TargetOpcode::BINARY_XOR => {
+        Mnemonic::BINARY_XOR => {
             apply_operator!("^");
         }
-        TargetOpcode::BINARY_AND => {
+        Mnemonic::BINARY_AND => {
             apply_operator!("&");
         }
-        TargetOpcode::BINARY_OR | TargetOpcode::INPLACE_OR => {
+        Mnemonic::BINARY_OR | Mnemonic::INPLACE_OR => {
             apply_operator!("|");
         }
-        TargetOpcode::UNARY_NOT => {
+        Mnemonic::UNARY_NOT => {
             apply_unary_operator!(!);
         }
-        TargetOpcode::UNARY_NEGATIVE => {
+        Mnemonic::UNARY_NEGATIVE => {
             apply_unary_operator!(-);
         }
-        TargetOpcode::BINARY_RSHIFT => {
+        Mnemonic::BINARY_RSHIFT => {
             let (tos, tos_accesses) = stack.pop().unwrap();
             let tos_value = tos.map(|tos| match tos {
                 Obj::Long(l) => Arc::clone(&l),
@@ -1027,7 +1021,7 @@ where
                 stack.push((None, tos_accesses));
             }
         }
-        TargetOpcode::BINARY_LSHIFT => {
+        Mnemonic::BINARY_LSHIFT => {
             let (tos, tos_accesses) = stack.pop().unwrap();
             let tos_value = tos.map(|tos| match tos {
                 Obj::Long(l) => Arc::clone(&l),
@@ -1055,7 +1049,7 @@ where
                 stack.push((None, tos_accesses));
             }
         }
-        TargetOpcode::LIST_APPEND => {
+        Mnemonic::LIST_APPEND => {
             let (tos, tos_modifiers) = stack.pop().unwrap();
             let tos_value = tos.map(|tos| {
                 match tos {
@@ -1088,7 +1082,7 @@ where
                 }
             }
         }
-        TargetOpcode::UNPACK_SEQUENCE => {
+        Mnemonic::UNPACK_SEQUENCE => {
             let (tos, tos_modifiers) = stack.pop().unwrap();
 
             tos_modifiers.push(access_tracking);
@@ -1109,7 +1103,7 @@ where
                 }
             }
         }
-        TargetOpcode::BUILD_SET => {
+        Mnemonic::BUILD_SET => {
             let mut set = std::collections::HashSet::new();
             let mut push_none = false;
 
@@ -1139,7 +1133,7 @@ where
                 ));
             }
         }
-        TargetOpcode::BUILD_TUPLE => {
+        Mnemonic::BUILD_TUPLE => {
             let mut tuple = Vec::new();
             let mut push_none = false;
 
@@ -1164,7 +1158,7 @@ where
                 stack.push((Some(Obj::Tuple(Arc::new(tuple))), tuple_accessors));
             }
         }
-        TargetOpcode::BUILD_MAP => {
+        Mnemonic::BUILD_MAP => {
             let tracking = InstructionTracker::new();
             tracking.push(access_tracking);
 
@@ -1174,7 +1168,7 @@ where
 
             stack.push((map, tracking));
         }
-        TargetOpcode::LOAD_GLOBAL => {
+        Mnemonic::LOAD_GLOBAL => {
             let tracking = InstructionTracker::new();
             tracking.push(access_tracking);
 
@@ -1183,20 +1177,20 @@ where
 
             stack.push((None, tracking));
         }
-        TargetOpcode::STORE_GLOBAL => {
+        Mnemonic::STORE_GLOBAL => {
             let (tos, accessing_instrs) = stack.pop().unwrap();
             let name = &code.names[instr.arg.unwrap() as usize];
             accessing_instrs.push(access_tracking);
             // Store TOS in a var slot
             globals.insert(Arc::clone(name), (tos, accessing_instrs));
         }
-        TargetOpcode::LOAD_DEREF => {
+        Mnemonic::LOAD_DEREF => {
             let tracking = InstructionTracker::new();
             tracking.push(access_tracking);
 
             stack.push((None, tracking));
         }
-        TargetOpcode::BUILD_LIST => {
+        Mnemonic::BUILD_LIST => {
             let mut list = Vec::new();
             // TODO: this is always true right now to avoid
             // testing empty sets that are added to as truthy values
@@ -1226,7 +1220,7 @@ where
                 ));
             }
         }
-        TargetOpcode::BUILD_CLASS => {
+        Mnemonic::BUILD_CLASS => {
             let (_tos, tos_accesses) = stack.pop().unwrap();
             let (_tos1, tos1_accesses) = stack.pop().unwrap();
             let (_tos2, tos2_accesses) = stack.pop().unwrap();
@@ -1236,21 +1230,21 @@ where
 
             stack.push((None, tos_accesses));
         }
-        TargetOpcode::MAKE_FUNCTION => {
+        Mnemonic::MAKE_FUNCTION => {
             let (_tos, tos_modifiers) = stack.pop().unwrap();
             let tos_modifiers = tos_modifiers.deep_clone();
             tos_modifiers.push(access_tracking);
 
             stack.push((None, tos_modifiers));
         }
-        TargetOpcode::POP_TOP => {
+        Mnemonic::POP_TOP => {
             let (_tos, tos_modifiers) = stack.pop().unwrap();
             tos_modifiers.push(access_tracking);
         }
-        TargetOpcode::GET_ITER => {
+        Mnemonic::GET_ITER => {
             // nop
         }
-        TargetOpcode::CALL_FUNCTION => {
+        Mnemonic::CALL_FUNCTION => {
             let accessed_instrs = InstructionTracker::new();
 
             let kwarg_count = (instr.arg.unwrap() >> 8) & 0xFF;
@@ -1291,7 +1285,7 @@ where
             //     stack[stack.len() - (1 + instr.arg.unwrap()) as usize]
             // );
         }
-        TargetOpcode::CALL_FUNCTION_VAR => {
+        Mnemonic::CALL_FUNCTION_VAR => {
             let (_additional_positional_args, arg_accesses) = stack.pop().unwrap();
             let accessed_instrs = arg_accesses.deep_clone();
 
@@ -1325,7 +1319,7 @@ where
 
             stack.push((result, accessed_instrs));
         }
-        TargetOpcode::CALL_FUNCTION_KW => {
+        Mnemonic::CALL_FUNCTION_KW => {
             let (_additional_kw_args, arg_accesses) = stack.pop().unwrap();
             let accessed_instrs = arg_accesses.deep_clone();
 
@@ -1359,7 +1353,7 @@ where
 
             stack.push((result, accessed_instrs));
         }
-        TargetOpcode::CALL_FUNCTION_VAR_KW => {
+        Mnemonic::CALL_FUNCTION_VAR_KW => {
             let (_additional_kw_args, arg_accesses) = stack.pop().unwrap();
             let accessed_instrs = arg_accesses.deep_clone();
             let (_additional_positional_args, arg_accesses) = stack.pop().unwrap();
@@ -1395,30 +1389,92 @@ where
 
             stack.push((result, accessed_instrs));
         }
-        TargetOpcode::POP_BLOCK | TargetOpcode::JUMP_ABSOLUTE => {
+        Mnemonic::POP_BLOCK | Mnemonic::JUMP_ABSOLUTE => {
             // nops
         }
-        TargetOpcode::PRINT_ITEM => {
+        Mnemonic::PRINT_ITEM => {
             stack.pop();
         }
-        TargetOpcode::PRINT_ITEM_TO => {
+        Mnemonic::PRINT_ITEM_TO => {
             stack.pop();
             stack.pop();
         }
-        TargetOpcode::PRINT_NEWLINE => {
+        Mnemonic::PRINT_NEWLINE => {
             // nop
         }
-        TargetOpcode::PRINT_NEWLINE_TO => {
+        Mnemonic::PRINT_NEWLINE_TO => {
             stack.pop();
         }
-        TargetOpcode::STORE_MAP => {
+        Mnemonic::STORE_MAP => {
             let (key, key_accesses) = stack.pop().unwrap();
             let (value, value_accesses) = stack.pop().unwrap();
             let (dict, dict_accesses) = stack.pop().unwrap();
-            panic!("{:?}{:?}{:?}", dict, key, value);
+
+            let mut new_accesses = dict_accesses;
+            new_accesses.extend(&value_accesses);
+            new_accesses.extend(&key_accesses);
+            new_accesses.push(access_tracking);
+
+            if dict.is_none() || key.is_none() || value.is_none() {
+                // We cannot track the state of at least one of these variables. Corrupt
+                // the entire state.
+                // TODO: this is a bit aggressive. In the future when we develop a new map type
+                // we should be able to track individual keys
+                stack.push((None, new_accesses));
+
+                return Ok(());
+            }
+
+            let dict_lock = dict.unwrap().extract_dict().unwrap();
+            let mut dict = dict_lock.write().unwrap();
+            let hashable_key: ObjHashable = key
+                .as_ref()
+                .unwrap()
+                .try_into()
+                .expect("key is not hashable");
+            dict.insert(hashable_key, value.unwrap());
+
+            drop(dict);
+
+            stack.push((Some(Obj::Dict(dict_lock)), new_accesses));
+        }
+        Mnemonic::MAP_ADD => {
+            return Err(
+                crate::error::ExecutionError::UnsupportedOpcode(Mnemonic::MAP_ADD.into()).into(),
+            );
+
+            // let (value, value_accesses) = stack.pop().unwrap();
+            // let (dict, dict_accesses) = stack.pop().unwrap();
+
+            // let mut new_accesses = dict_accesses;
+            // new_accesses.extend(&value_accesses);
+            // new_accesses.push(access_tracking);
+
+            // if dict.is_none() || key.is_none() || value.is_none() {
+            //     // We cannot track the state of at least one of these variables. Corrupt
+            //     // the entire state.
+            //     // TODO: this is a bit aggressive. In the future when we develop a new map type
+            //     // we should be able to track individual keys
+            //     stack.push((None, new_accesses));
+
+            //     return Ok(());
+            // }
+
+            // let arc_dict = dict.unwrap().extract_dict().unwrap();
+            // let mut dict = arc_dict.write().unwrap();
+            // let hashable_key: ObjHashable = key.as_ref().unwrap().try_into().expect("key is not hashable");
+            // dict.insert(hashable_key, value.unwrap());
+
+            // drop(dict);
+
+            // stack.push((Some(Obj::Dict(arc_dict)), new_accesses));
+        }
+        Mnemonic::YIELD_VALUE => {
+            // todo: add to generator
+            let (_tos, _accesses) = stack.pop().unwrap();
         }
         other => {
-            return Err(crate::error::ExecutionError::UnsupportedOpcode(other).into());
+            return Err(crate::error::ExecutionError::UnsupportedOpcode(other.into()).into());
         }
     }
 
@@ -1427,14 +1483,14 @@ where
 
 /// Represents an instruction that was parsed from its raw bytecode.
 #[derive(Debug, Clone, Eq, PartialEq)]
-pub enum ParsedInstr {
-    Good(Arc<Instruction<TargetOpcode>>),
+pub enum ParsedInstr<O: Opcode<Mnemonic = py27::Mnemonic>> {
+    Good(Arc<Instruction<O>>),
     Bad,
 }
 
-impl ParsedInstr {
+impl<O: Opcode<Mnemonic = py27::Mnemonic>>  ParsedInstr<O> {
     #[track_caller]
-    pub fn unwrap(&self) -> Arc<Instruction<TargetOpcode>> {
+    pub fn unwrap(&self) -> Arc<Instruction<O>> {
         if let ParsedInstr::Good(ins) = self {
             Arc::clone(ins)
         } else {
@@ -1447,18 +1503,18 @@ impl ParsedInstr {
 /// codepaths. This will only decode instructions that are either proven statically
 /// to be taken (with `JUMP_ABSOLUTE`, `JUMP_IF_TRUE` with a const value that evaluates
 /// to true, etc.)
-pub fn const_jmp_instruction_walker<F>(
+pub fn const_jmp_instruction_walker<F, O: Opcode<Mnemonic = py27::Mnemonic>>(
     bytecode: &[u8],
     consts: Arc<Vec<Obj>>,
     mut callback: F,
-) -> Result<BTreeMap<u64, ParsedInstr>, Error>
+) -> Result<BTreeMap<u64, ParsedInstr<O>>, Error<O>>
 where
-    F: FnMut(&Instruction<TargetOpcode>, u64) -> WalkerState,
+    F: FnMut(&Instruction<O>, u64) -> WalkerState,
 {
     let debug = !true;
     let mut rdr = Cursor::new(bytecode);
     let mut instruction_sequence = Vec::new();
-    let mut analyzed_instructions = BTreeMap::<u64, ParsedInstr>::new();
+    let mut analyzed_instructions = BTreeMap::<u64, ParsedInstr<O>>::new();
     // Offset of instructions that need to be read
     let mut instruction_queue = VecDeque::<u64>::new();
 
@@ -1567,8 +1623,8 @@ where
 
         if instr.opcode.is_jump() {
             if matches!(
-                instr.opcode,
-                TargetOpcode::JUMP_ABSOLUTE | TargetOpcode::JUMP_FORWARD
+                instr.opcode.mnemonic(),
+                Mnemonic::JUMP_ABSOLUTE | Mnemonic::JUMP_FORWARD
             ) {
                 // We've reached an unconditional jump. We need to decode the target
                 let target = if instr.opcode.is_relative_jump() {
@@ -1585,7 +1641,7 @@ where
                 }
 
                 rdr.set_position(target);
-                match decode_py27(&mut rdr) {
+                match decode_py27::<O, _>(&mut rdr) {
                     Ok(_instr) => {
                         // Queue the target
                         queue!(target, state.force_queue_next());
@@ -1627,7 +1683,7 @@ where
             }
         }
 
-        if instr.opcode != TargetOpcode::RETURN_VALUE && instr.opcode != TargetOpcode::RAISE_VARARGS
+        if instr.opcode.mnemonic() != Mnemonic::RETURN_VALUE && instr.opcode.mnemonic() != Mnemonic::RAISE_VARARGS
         {
             queue!(next_instr_offset, state.force_queue_next());
         }
@@ -1640,9 +1696,9 @@ where
     Ok(analyzed_instructions)
 }
 
-fn remove_bad_instructions_behind_offset(
+fn remove_bad_instructions_behind_offset<O: Opcode<Mnemonic = py27::Mnemonic>>(
     offset: u64,
-    analyzed_instructions: &mut BTreeMap<u64, Arc<Instruction<TargetOpcode>>>,
+    analyzed_instructions: &mut BTreeMap<u64, Arc<Instruction<O>>>,
 ) {
     // We need to remove all instructions parsed between the last
     // conditional jump and this instruction
@@ -1671,21 +1727,21 @@ fn remove_bad_instructions_behind_offset(
     }
 }
 
-    #[macro_export]
-    macro_rules! Instr {
-        ($opcode:expr) => {
-            Instruction {
-                opcode: $opcode,
-                arg: None,
-            }
-        };
-        ($opcode:expr, $arg:expr) => {
-            Instruction {
-                opcode: $opcode,
-                arg: Some($arg),
-            }
-        };
-    }
+#[macro_export]
+macro_rules! Instr {
+    ($opcode:expr) => {
+        Instruction {
+            opcode: $opcode,
+            arg: None,
+        }
+    };
+    ($opcode:expr, $arg:expr) => {
+        Instruction {
+            opcode: $opcode,
+            arg: Some($arg),
+        }
+    };
+}
 
 #[cfg(test)]
 pub(crate) mod tests {
@@ -1693,9 +1749,9 @@ pub(crate) mod tests {
     use num_bigint::BigInt;
     use py27_marshal::bstr::BString;
 
-    use std::sync::Arc;
+    use std::sync::{Arc, RwLock};
 
-    type TargetOpcode = pydis::opcode::Python27;
+    type TargetOpcode = pydis::opcode::py27::Standard;
 
     #[macro_export]
     macro_rules! Long {
@@ -1704,6 +1760,13 @@ pub(crate) mod tests {
         };
     }
 
+    #[macro_export]
+    macro_rules! String {
+        ($value:expr) => {
+            py27_marshal::Obj::String(Arc::new(bstr::BString::from($value)))
+        };
+    }
+
     #[test]
     fn binary_xor() {
         let (mut stack, mut vars, mut names, mut globals, names_loaded) = setup_vm_vars();
@@ -2216,6 +2279,188 @@ pub(crate) mod tests {
         }
     }
 
+    #[test]
+    fn store_subscr_list() {
+        let (mut stack, mut vars, mut names, mut globals, names_loaded) = setup_vm_vars();
+        let mut code = default_code_obj();
+
+        let key = Long!(0);
+        let value = Long!(0x41);
+
+        let mut expected_list = vec![0x41];
+
+        let actual_list = Obj::List(Arc::new(RwLock::new(vec![Long!(0)])));
+        let consts = vec![actual_list.clone(), key, value];
+
+        Arc::get_mut(&mut code).unwrap().consts = Arc::new(consts);
+
+        let instrs = [
+            // Load value on to stack
+            Instr!(TargetOpcode::LOAD_CONST, 2),
+            // Load list on to stack
+            Instr!(TargetOpcode::LOAD_CONST, 0),
+            // Load key on to stack
+            Instr!(TargetOpcode::LOAD_CONST, 1),
+            Instr!(TargetOpcode::STORE_SUBSCR),
+        ];
+
+        for instr in &instrs {
+            execute_instruction(
+                instr,
+                Arc::clone(&code),
+                &mut stack,
+                &mut vars,
+                &mut names,
+                &mut globals,
+                Arc::clone(&names_loaded),
+                |_f, _args, _kwargs| {
+                    panic!("functions should not be invoked");
+                },
+                (),
+            )
+            .expect("unexpected error")
+        }
+
+        assert!(stack.is_empty());
+
+        match &actual_list {
+            Obj::List(list_lock) => {
+                let list = list_lock.read().unwrap();
+                assert_eq!(list.len(), 1);
+
+                assert_eq!(*list[0].clone().extract_long().unwrap(), BigInt::from(0x41));
+            }
+            other => panic!("unexpected type: {:?}", other.typ()),
+        }
+    }
+
+    #[test]
+    fn store_subscr_dict() {
+        let (mut stack, mut vars, mut names, mut globals, names_loaded) = setup_vm_vars();
+        let mut code = default_code_obj();
+
+        let key = String!("key");
+        let value = Long!(0x41);
+
+        let mut expected_hashmap = HashMap::new();
+        expected_hashmap.insert(ObjHashable::try_from(&key).unwrap(), value.clone());
+
+        let actual_dict = Obj::Dict(Default::default());
+        let consts = vec![actual_dict.clone(), key, value];
+
+        Arc::get_mut(&mut code).unwrap().consts = Arc::new(consts);
+
+        let instrs = [
+            // Load value on to stack
+            Instr!(TargetOpcode::LOAD_CONST, 2),
+            // Load dict on to stack
+            Instr!(TargetOpcode::LOAD_CONST, 0),
+            // Load key on to stack
+            Instr!(TargetOpcode::LOAD_CONST, 1),
+            Instr!(TargetOpcode::STORE_SUBSCR),
+        ];
+
+        for instr in &instrs {
+            execute_instruction(
+                instr,
+                Arc::clone(&code),
+                &mut stack,
+                &mut vars,
+                &mut names,
+                &mut globals,
+                Arc::clone(&names_loaded),
+                |_f, _args, _kwargs| {
+                    panic!("functions should not be invoked");
+                },
+                (),
+            )
+            .expect("unexpected error")
+        }
+
+        assert!(stack.is_empty());
+
+        match &actual_dict {
+            Obj::Dict(dict_lock) => {
+                let actual_dict = dict_lock.read().unwrap();
+                for (key, expected_value) in &expected_hashmap {
+                    let actual_value = actual_dict.get(key);
+
+                    assert!(actual_value.is_some());
+
+                    let actual_value = actual_value.unwrap().clone().extract_long();
+                    let expected_value = expected_value.clone().extract_long().unwrap();
+
+                    assert_eq!(expected_value, actual_value.unwrap());
+                }
+            }
+            other => panic!("unexpected type: {:?}", other.typ()),
+        }
+    }
+
+    #[test]
+    fn store_map() {
+        let (mut stack, mut vars, mut names, mut globals, names_loaded) = setup_vm_vars();
+        let mut code = default_code_obj();
+
+        let key = String!("key");
+        let value = Long!(0x41);
+
+        let mut expected_hashmap = HashMap::new();
+        expected_hashmap.insert(ObjHashable::try_from(&key).unwrap(), value.clone());
+
+        let consts = vec![Obj::Dict(Default::default()), key, value];
+
+        Arc::get_mut(&mut code).unwrap().consts = Arc::new(consts);
+
+        let instrs = [
+            // Load dict on to stack
+            Instr!(TargetOpcode::LOAD_CONST, 0),
+            // Load value on to stack
+            Instr!(TargetOpcode::LOAD_CONST, 2),
+            // Load key on to stack
+            Instr!(TargetOpcode::LOAD_CONST, 1),
+            Instr!(TargetOpcode::STORE_MAP),
+        ];
+
+        for instr in &instrs {
+            execute_instruction(
+                instr,
+                Arc::clone(&code),
+                &mut stack,
+                &mut vars,
+                &mut names,
+                &mut globals,
+                Arc::clone(&names_loaded),
+                |_f, _args, _kwargs| {
+                    panic!("functions should not be invoked");
+                },
+                (),
+            )
+            .expect("unexpected error")
+        }
+
+        // The dict should still be on the stack
+        assert_eq!(stack.len(), 1, "stack size is not 1");
+
+        match &stack[0].0 {
+            Some(Obj::Dict(dict)) => {
+                let actual_dict = dict.read().unwrap();
+                for (key, expected_value) in &expected_hashmap {
+                    let actual_value = actual_dict.get(key);
+
+                    assert!(actual_value.is_some());
+
+                    let actual_value = actual_value.unwrap().clone().extract_long();
+                    let expected_value = expected_value.clone().extract_long().unwrap();
+
+                    assert_eq!(expected_value, actual_value.unwrap());
+                }
+            }
+            Some(other) => panic!("unexpected type: {:?}", other.typ()),
+            _ => panic!("unexpected None value for TOS"),
+        }
+    }
+
     pub(crate) fn setup_vm_vars() -> (
         VmStack<()>,
         VmVars<()>,
